#include <xc.h>
#include "interrupts.h"
#include "common.h"
#include "commands.h"

#define MaxSSTCycles 5                                                          // Max cycles for SST39SF040 Jedec Command

const unsigned char SSTCmdPages[MaxSSTCycles]={1,0,1,1,0};                      // Page for JEDEC Command 5555h and 2222h for 3 and 5 pre-sequences

unsigned char TMR1Cnt;                                                          // Timer1 SW PostScaler Counter (Timer 1 is used for Double Click detection)   
unsigned char TMR2Cnt;                                                          // Timer2 SW PostScaler Counter (Timer 2 is used for LongPress detection)
unsigned char TMR1PS = TMR1SWPostScaler;                                        // Timer1 SW Postscaler preset

bool Switch1PressedFlag = false;                                                // Software Flag - Switch push - Not directly used
bool Switch1ReleasedFlag = false;                                               // Software Flag - Switch release - Not directly used
bool WasLockedFlag = false;                                                     // Software Flag - A command 46 is being received while coming from a Locked Status

extern void Select_Eeprom_Bank(unsigned char BankN);
extern void Restore_Internal_ROM(bool reset);
extern void Enable_Dandanator_Commands();

/*************************************************************************
* ISR Called: Launch Timer for Long Press Detection
*************************************************************************/
void ISR_Launch_Long_Press_Timer(void)
{  
  TMR2   = TMR2InitValue;                                                       // Load TMR2 Counter for ~8ms @32Mhz
  TMR2Cnt = TMR2SWPostScaler;                                                   // Load SW Postscaler Value
  TMR2IF  = 0;                                                                  // Clear TMR4 Interrupt Flag
  LongSwitchPressedFlag = false;                                                // Clear LongPress Flag
  TMR2IE  = 1;                                                                  // Enable TMR2 1-Shot Interrupt
}  

/*************************************************************************
* ISR Called: Launch Timer for Double Click Detection
*************************************************************************/
void ISR_Launch_Double_Press_Timer(void)
{  
  TMR1H=TMR1InitValueH;                                                         // Set for 50ms int @32mhz and prescaler 1:8
  TMR1L=TMR1InitValueL;
  TMR1Cnt=TMR1PS;                                                               // Prepare software postscaler for action after ~200ms (default)
  TMR1IF  = 0;                                                                  // Clear TMR1 Interrupt Flag
  DoubleSwitchPressedFlag = false;                                              // ClearDoublePress Flag
  TMR1IE  = 1;                                                                  // Enable TMR1 1-Shot Interrupt
}  

/*************************************************************************
* ISR Called: Launch Timer for Spectrum Commands timeout
*************************************************************************/
ISR_Launch_Spectrum_Incoming_Timeout_Timer(unsigned char TMRVal)
{  
  TMR0   = TMRVal;                                                              // Load TMR0 Counter for timeout
  TMR0IF  = 0;                                                                  // Clear TMR0 Interrupt Flag
  TMR0IE  = 1;                                                                  // Enable TMR0 1-Shot Interrupt
}  


/*************************************************************************
* ISR Called: Process SST Command driven from Spectrum
*************************************************************************/
void ISR_SST_Command(unsigned char SSTCycles, unsigned char SlotN, unsigned char SavedLata)
{
unsigned char i;
  for (i=0;i!=SSTCycles;i++) 
   {
     LATA=SSTCmdPages[i];                                                       // Select Slot depending on JEDEC command
     INTF=0;
     while(INTF==0);                                                            // Wait write operation from Spectrum
   }
   LATA=SlotN;
   INTF=0;
   while(INTF==0);                                                              // Wait write operation from Spectrum
  // LATA=SavedLata;                                                            // DEBUG : , do not return to lata state
   INTF=0;
}

/*************************************************************************
* ISR Called: Execute Fast Command
*************************************************************************/
ISR_Execute_Fast_Command()
{  
#define CMD_Lock 1                                                              // Subcommands of command 40
#define CMD_Enable 16
#define CMD_Disable 31    
#define CMD_Reboot 40
#define CMD_EEP_ProgramByte 1                                                   // Subcommands of command 48
#define CMD_EEP_EraseSector 16
#define CMD_EEP_ProgramSector 32
#define EEPBLKSIZE 4096                                                         // SST39SF040 sector size              
    
unsigned char SWAction;
unsigned char SWBehaviour;
unsigned char EepPage;                                                          // Calculated page for external eeprom Programming operation
unsigned char CurLATA;                                                          // Current LATA Value
unsigned short bytecnt;                                                         // Counter for sector write
bool ResetReq;                                                                  // Reset Requested

  switch (SpectrumCommandCounter)
  {
      case CMD_SP_FASTCHANGE:
      {                             
          ResetReq=((SpectrumData2Counter&0x01)==0x01);
          if  (ResetReq)
          {
              Z80_nReset_Hold();
              ChangeSlotOnReset=false;                                          // Clear Change Slot on Spectrum-Driven-Reset
          }
          if (SpectrumData1Counter==PreviousROMBank)                            // Modify target slot if "previous slot" was selected
            SpectrumData1Counter=PreviousSlot;
          if (SpectrumData1Counter==InternalROMBank)
          {
            Restore_Internal_ROM(false);
          }
          else
          {
            Select_Eeprom_Bank(SpectrumData1Counter-1);                         // Warning : It will not page in external rom 
          }
          if ((SpectrumData2Counter&0x08)==0x08) 
          {
            Disable_Dandanator_Commands();  
          }
          else
          {
            if ((SpectrumData2Counter&0x04)==0x04) 
           
            {
              SpectrumCommandStatus=Commands_Locked;   
            }
            else
            {
              SpectrumCommandStatus=Wait_Normal;
            }
          }
          if (ResetReq)
          {
             Z80_nReset_Release();
          }
          if ((SpectrumData2Counter&0x02)==0x02) 
          {
            Z80_Send_NMI();   
          }
          break;
      }
      case CMD_SP_STOREBOOTBANK:
      {
          ModifyEepromFlag = true;                                              // Eeprom must be modified from main thread. Flag, Address & value
          EepromAddress = EE_NormalBoot+SpectrumData2Counter;
          ModifyEepromValue = SpectrumData1Counter;
          Disable_Dandanator_Commands();                                        // Disable Commands until action is performed
          break;
      }
      case CMD_SP_SW1MODE:
      {
          SWAction=(SpectrumData2Counter>>3)&3;                                 // Get bits 4..3 value
          SWBehaviour=SpectrumData2Counter&7;                                   // Get bits 2..0 value
          if (SWAction!=3)                                                      // 3 is for adjusting double click detecting window
          {
            Switches[SW1][SWAction].action=SWBehaviour;                         // Modify Switch 1 behaviour
            Switches[SW1][SWAction].BankN=SpectrumData1Counter;
          }
          else
          {
            TMR1PS=SWBehaviour+SpectrumData1Counter;                            // Update number of 50ms postscaler for double click window allowance                      
          }
          SpectrumCommandStatus=Wait_Normal;
          break;
      }
      case CMD_SP_SW2MODE:
      {
          SWAction=(SpectrumData2Counter>>3)&3;                                 // Get bits 4..3 value
          SWBehaviour=SpectrumData2Counter&7;                                   // Get bits 2..0 value
          Switches[SW2][SWAction].action=SWBehaviour;                           // Modify Switch 2 behaviour
          Switches[SW2][SWAction].BankN=SpectrumData1Counter;    
          SpectrumCommandStatus=Wait_Normal;
          break;
      }
      case CMD_SP_STOREDATA:
      {
          ModifyEepromFlag = true;                                              // Eeprom must be modified from main thread. Flag, Address & value
          EepromAddress = SpectrumData1Counter;
          ModifyEepromValue = SpectrumData2Counter;
          Disable_Dandanator_Commands();                                        // Disable Commands until action is performed
          break;
      }  
      case CMD_SP_RETRIEVEDATA:
      {
          ReadEepromFlag = true;                                                // Eeprom must be read from main thread. Flag, Address & value
          EepromAddress = SpectrumData1Counter;
          ModifyEepromValue = SpectrumData2Counter;
          Disable_Dandanator_Commands();                                        // Disable Commands until action is performed
          break;
      }  
      case CMD_SP_LOCKUNLOCKDISABLE:
      {
        if (SpectrumData1Counter == SpectrumData2Counter)                       // Execute only if Data1==Data2
        {        
          switch(SpectrumData1Counter)
          {
            case CMD_Lock:
            {
                SpectrumCommandStatus=Commands_Locked;
                break;
            }
            case CMD_Disable:
            {
                Disable_Dandanator_Commands();
                break;
            }
            case CMD_Enable:
            {
                WasLockedFlag=false;
                Enable_Dandanator_Commands();                                   // DEBUG : Warning function called from two threads
                break;
            }
            default: 
            {
                SpectrumCommandStatus=Commands_Locked;
                break;
            }
          }
        }
        else
        {
                SpectrumCommandStatus=Commands_Locked;  
        }
        break;   
      }
      case CMD_SP_PICREBOOT:
      {
          if (SpectrumData1Counter == SpectrumData2Counter)                     // Execute only if Data1==Data2)
              if (SpectrumData1Counter == CMD_Reboot)
              {
                WDTCONbits.SWDTEN=1;                                            // Enable watch Dog to reset PIC
                while(true);
              }
          break;
      }
      case CMD_EEP_PROGRAM:                                                     // WARNING: External Eeprom MUST be paged in. Also this is a deadlock if
      {                                                                         // The Spectrum fails to follow the sequence
          EepPage=(SpectrumData2Counter>>2);                                    // Calculate page (4k sectors) and set CE to 0
          CurLATA=LATA;                                                         // Save Current LATA

          if (SpectrumData1Counter == CMD_EEP_EraseSector)                      // If sector erase requested
          {
            ISR_SST_Command(5,EepPage,CurLATA);
          }
          else
          {
            if (SpectrumData1Counter == CMD_EEP_ProgramByte)                    // If Program Byte requested
            {
              ISR_SST_Command(3,EepPage,CurLATA);   
            } 
            else
              if (SpectrumData1Counter == CMD_EEP_ProgramSector)                // If Program Sector requested
              {
               for (bytecnt=0;bytecnt!=EEPBLKSIZE;bytecnt++)                    // Enter a loop of byte program
                 ISR_SST_Command(3,EepPage,CurLATA);   
              } 
          }
          SpectrumCommandStatus=Wait_Normal;
          break;
      }
      case CMD_SP_CMDALLOW:
      {
          ModifyEepromFlag = true;                                              // Eeprom must be modified from main thread. Flag, Address & value
          EepromAddress = SpectrumData2Counter-1+EE_ExtraRom_CMDAllowed1;
          ModifyEepromValue = SpectrumData1Counter;
          Disable_Dandanator_Commands();                                        // Disable Commands until action is performed
          break;
      }
      default: break;
  }
}

/*************************************************************************
* Interrupt Handling Routine
*************************************************************************/
void  Interrupt_Handler()    @0x0204                                            // Interrupt Service Routine Called from bootloader
{
  // INT - Write Request to ROM - Commands
  //----------------------------------
  if ((INTF==1) && (INTE == 1))                                                 // Request to write in ROM from Spectrum (Used for automated bank Change)
  {
    switch (SpectrumCommandStatus)
    {
      case Wait_Normal:
      {
        SpectrumCommandStatus=In_Progress_Normal;                               // Change State Machine
        SpectrumCommandCounter=1;                                               // Init Command Counter
        ISR_Launch_Spectrum_Incoming_Timeout_Timer(TMR0InitValue);              // Launch timeout timer to decide Command ReadyFlag
        break;
      }
      case In_Progress_Normal:
      {
        SpectrumCommandCounter++;                                               // Go on counting command pulses  
        TMR0=TMR0InitValue;                                                     // Reset Timeout timer
        break;
      }
      case In_Progress_Special_Data1:
      {
        SpectrumData1Counter++;                                                 // Go on counting Data 1 pulses  
        TMR0=TMR0InitValue;                                                     // Reset Timeout timer
        break;
      }
      case In_Progress_Special_Data2:
      {
        SpectrumData2Counter++;                                                 // Go on counting Data 2 pulses  
        TMR0=TMR0InitValue;                                                     // Reset Timeout timer
        break;
      }
      case In_Progress_Fast:
      {            
        TMR0IE=0;                                                               // Disable timeout detection
        ISR_Execute_Fast_Command();                                             // Fast execution of last command
        break;  
      }
      case Commands_Locked:                                                     // Try to find an Unlock Command
      {
        SpectrumCommandStatus=In_Progress_Locked;                               // Change State Machine
        SpectrumCommandCounter=1;                                               // Init Command Counter
        ISR_Launch_Spectrum_Incoming_Timeout_Timer(TMR0InitValue);              // Launch Timeout timer to decide if Unlock Command;
        break;
      }
      case In_Progress_Locked:
      {
        SpectrumCommandCounter++;                                               // Go on Counting command pulses 
        TMR0=TMR0InitValue;                                                     // Reset Timeout timer
        break;
      }
      default: break;
    }
    INTF = 0;                                                                   // Clear interrupt Flag
  }

  // IOC - Switch
  //----------------------------------
  #define ButtonDebounce_ms 5                                                  // Button Debounce in milliseconds
  if ((IOCIF==1) && (IOCIE == 1))                                               
  {
    __delay_ms(ButtonDebounce_ms);                                              // Debounce button(s)
    if (IOCBFbits.IOCBF1 == 1)                                                  // IOCF Flag For RB1: Switch 1
    {      
      if (Switch1 == 0)                                                         // Button1 Depressed
      { 
        if (Switch1ReleasedFlag)                                                // If previous press not cleared 
        {
          Switch1ReleasedFlag = false;                                          // Clear previous release
          Switch1PressedFlag = false;                                           // Clear previous press
          LongSwitchPressedFlag = false;                                        // Clear LongSwitchPresedFlag
          ShortSwitchPressedFlag = false;                                       // Clear ShortSwitchPressedFlag
          DoubleSwitchPressedFlag = true;                                       // Set Double Switch Press Flag 
          TMR1IE = 0;                                                           // Disable Double click detection
        }
        else
        {
          Switch1PressedFlag = true;
          ISR_Launch_Long_Press_Timer();                                        // Launch Timer to detect LongPress
        }
      }
      else
      {
        if (Switch1PressedFlag)                                                 // If switch is still depressed (not cleared by long press)
        {
          Switch1ReleasedFlag = true;                                           // Signal button was released after a short time
          Switch1PressedFlag = false;                                           // Clear Pressed Flag
          LongSwitchPressedFlag = false;                                        // Clear LongSwitchPresedFlag
          ShortSwitchPressedFlag = false;                                       // Clear ShortSwitchPressedFlag
          DoubleSwitchPressedFlag = false;                                      // Clear Double Switch Press Flag
          ISR_Launch_Double_Press_Timer();                                      // Launch Double Switch
          TMR2IE = 0;                                                           // Disable LongPress Detection Timer
        }
      }
      IOCBF&=4;                                                                 // Clear all bits but bit2 (corresponding to RB2 - Switch 2)
    }
    if (IOCBFbits.IOCBF2 == 1)                                                  // IOCF Flag For RB2: Switch 2
    {    
      if (Switch2==0)                                                           // DEBUG - Apparently IOCBP is not working as it should
        Switch2PressedFlag=true;
      IOCBF&=2;                                                                 // Clear all bits but bit1 (corresponding to RB1 - Switch 1)
    }
                       
  
  }
  
    // TMR1 - Double Click Detection
  //-------------------------------------------
  if ((TMR1IF==1) && (TMR1IE==1))                                               // TMR1 Interrupt
  {
    TMR1Cnt--;                                                                  // Decrement SW PostScaler
    if (TMR1Cnt == 0)                                                           // Action Trigger
    {
      Switch1ReleasedFlag = false;                                              // Clear previous release
      Switch1PressedFlag = false;                                               // Clear previous press
      LongSwitchPressedFlag = false;                                            // Clear LongSwitchPresedFlag
      ShortSwitchPressedFlag = true;                                            // Set ShortSwitchPressedFlag
      DoubleSwitchPressedFlag = false ;                                         // Clear Double Switch Press Flag 
      TMR1IE = 0;                                                               // Disable Double click detection
    }
    else                                                                        // Execute Action
    {
      TMR1H=TMR1InitValueH;                                                     // Set for 50ms int @32mhz and prescaler 1:8
      TMR1L=TMR1InitValueL;
    }
    TMR1IF = 0;                                                                 // Clear Interrupt Flag
  }
  
  // TMR2 - Long Press Detection Timer - 1-Shot 
  //-------------------------------------------
   if ((TMR2IF==1) && (TMR2IE==1))                                              // TMR2 Interrupt
  {
    TMR2Cnt--;                                                                  // Decrement SW PostScaler
    if (TMR2Cnt == 0)                                                           // Action Trigger
    {
      LongSwitchPressedFlag = true;                                             // Signal Flag LongPress
      Switch1PressedFlag = false;                                               // Clear SwitchPressed Flag
      Switch1ReleasedFlag = false;                                              // Clear Potential SwitchReleased Flags
      ShortSwitchPressedFlag = false;                                           // Clear ShortSwitchPressedFlag
      DoubleSwitchPressedFlag = false ;                                         // Clear Double Switch Press Flag 
      TMR2IE = 0;                                                               // Disable TMR2 Interrupt
    }
    else
    {
      TMR2 = 255;                                                               // Load TMR2 to max
    }
    TMR2IF = 0;                                                                 // Clear Interrupt Flag
  }

  // TMR0 - Spectrum Command Timeout
  //--------------------------------------------------------
   if ((TMR0IF==1) && (TMR0IE==1))                                              // TMR0 Interrupt
  {

      switch (SpectrumCommandStatus)
      {       
        case In_Progress_Normal:    { // Normal or 1st stage special command was in progress and it has timed out
                                      if ((SpectrumCommandCounter>39)&&(SpectrumCommandCounter<50)) // if special command (40-49 Special commands)
                                      {
                                        SpectrumCommandStatus=In_Progress_Special_Data1;
                                        SpectrumData1Counter=0;
                                        ISR_Launch_Spectrum_Incoming_Timeout_Timer(TMR0SpecialAllowanceInitValue);  // Launch Command timer for receiving Data 1
                                      }
                                      else
                                      { 
                                        SpectrumCommandStatus=Pending_Processing;
                                        TMR0IE= 0;                              // Disable TMR0 
                                        INTE=0;                                 // Disable Command reception until command is consumed (without changing Status)
                                      }

                                      break;
                                    }
        case In_Progress_Locked:    { // Expecting Unlock Command is timed out
                                      if (SpectrumCommandCounter==CMD_SP_LOCKUNLOCKDISABLE) // if Unlock Special Command
                                      {
                                        WasLockedFlag=true;                     // Signal flag to complement In_Progress_Special_Data1 status
                                        SpectrumCommandStatus=In_Progress_Special_Data1;
                                        SpectrumData1Counter=0;
                                        ISR_Launch_Spectrum_Incoming_Timeout_Timer(TMR0SpecialAllowanceInitValue);  // Launch Command timer for receiving Data 1
                                      }
                                      else
                                      {
                                        SpectrumCommandStatus=Commands_Locked;  // Discard Command if not Special Unlock Command
                                      } 
                                      break;
                                    }
        case In_Progress_Special_Data1:{ // Special Command (second stage, Data 1 reception) was in progress and its window just ended
                                         SpectrumCommandStatus=In_Progress_Special_Data2;  
                                         SpectrumData2Counter=0;
                                         ISR_Launch_Spectrum_Incoming_Timeout_Timer(TMR0SpecialAllowanceInitValue); // Launch Command timer for receiving Data 2
                                         break;
                                       }
        case In_Progress_Special_Data2:{ // Special Command (third stage, Data 2 reception) was in progress and its window just ended
                                         SpectrumCommandStatus=In_Progress_Fast;  
                                         ISR_Launch_Spectrum_Incoming_Timeout_Timer(TMR0SpecialAllowanceInitValue); // Launch Command timer for receiving Fast execute Pulse
                                         break;
                                       }
        case In_Progress_Fast:      { // Special Command (third stage) was in progress and its window ended without confirmation - TIMEOUT
                                        if (WasLockedFlag)
                                        {
                                            WasLockedFlag=false;
                                            SpectrumCommandStatus=Commands_Locked;
                                        }
                                        else
                                        {
                                            SpectrumCommandStatus=Wait_Normal;
                                        }
                                      TMR0IE= 0;                                // Disable TMR0 
                                      break;
                                    }
      }                          
    TMR0IF = 0;                                                                 // Clear Interrupt Flag
  }
}
