/**********************************************************************
* 2016/2017  Dandare: Controller for ZX Dandanator! Mini
* Version:        	2017.09.22
* uController:     	PIC16F1826
* REVISION HISTORY:
*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
* Author        Date        Comments on this revision
*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
* Dandare       2016.06.09  Initial Release Adapted from PIC16F1825 Prototype
* Dandare       2016.06.10  Corrected Bug with launch of Extra ROM, Enable PWRTE, Code for ROMCS Request, Expanded Reset & Wait Time for better 128k compatibility.
* Dandare       2016.06.13  Changed for v3.x Romset Compatibility 
* Dandare       2016.06.14  Changed LongPress to also Reset PIC (deprecated 2016.06.23)
* Dandare       2016.06.17  Speed increased to 32 Mhz. ~12us bank change, Pic Reset and Rom launch adjusted with STATUS.nTO bit, NMI now allowed on Slots 0 and 31.
* Dandare       2016.06.20  Adaptation for usage-independent bank switching - Complete Protocol rework 
* Dandare       2016.06.21  Adaptation for usage-independent bank switching - in progress. Special Commands
* Dandare       2016.06.22  Adaptation for usage-independent bank switching - testing after Z80 Code adaptation
* Dandare       2016.06.23  Most Commands tested - Added double click detection window (50-350ms) as command. Adjusted Button debounce
* Dandare       2016.06.27  Small change to Command 40 and correction of Status "Commands Locked" ISR. Corrected Missing break on CMD_SP_LOCKUNLOCKDISABLE
* Dandare       2016.07.06  Buttons defaults changed. Command 41 Window duration changed to allow longer windows
* Dandare       2016.07.07  Normal Command 33 implemented. Tweaked boot. Solved Issue with Timer1 HL-pair register reload
* Dandare       2016.07.08  Adapted for use with a BootLoader - Minor tweaks to boot sequence to work alongside bootloader
* Dandare       2016.07.10  Further tweaks to minimize Spectrum Boot Delay 
* Dandare       2016.07.15  Small tweak to better detect Spectrum Type on boot if reset is pressed. This tweak also speeds up normal boot process.
* Dandare       2016.08.03  Corrections when using normal commands to return to external rom from internal ROM. Temporarily launch first button with commands enabled
* Dandare       2016.08.04  Added normal command 34 (go to internal rom & disable further commands). Change calls to slot 32, now called with commands locked by default.
* Dandare       2016.08.07  Added normal command 39 (Set slot to change to in case of Spectrum-Reset). Unlocked location for SpectrumType var.
* Dandare       2016.08.12  Disable reset slot if upstream RomCS is detected. Force +2/+3 RomCS (A25 Bottom 25) to support multiple dandanators also on +2/+3 
* Dandare       2016.08.20  Change communication protocol from window based to timeout based. Major Change. Not backwards compatible
* Dandare       2016.08.21  Now NMI action on button also enables commands. Changed NMI ROM behavior and Code
* Dandare       2016.09.08  Add Special Command 48 for programming external eeprom
* Dandare       2016.09.15  Add subcommand to command 48 for sector (4k) write.  
* Dandare       2016.09.16  Force boot into internal rom now enables commands but locks them (for TAP/Audio eeprom loader function)
* Dandare       2016.11.02  When detecting a conflict (upstream ROMCS) now it LOCKS the commands instead of disabling them (in preparation for Divide/DivMMC update)
* Dandare       2016.11.04  Modified Boot_Dandanator to check for RomCS or reset
* Dandare       2016.11.05  Yet another tweak to the boot process for easier upgrade from divide. Include eeprom initialization in HEX.
* Dandare       2016.11.07  And yet another tweak for the divide ROMCS/OE1 connection on +2a/+3 (divide jumper)
* Dandare       2016.11.08  And the final change for transparent +2/+3 divide upgrade
* Dandare       2016.11.18  Added "double click left button for next rom" functionality when in ZX Testing mode (will jump to previous slot, whatever it contains)
* Dandare       2017.03.05  Command timeouts adjusted to support CP/M in allram mode -> RELEASE 6.4
* Dandare       2017.05.04  Added slot 35 to special command 40 for returning to previous selected slot. Added Boot on Slot 2. Little code cleanup 
* Dandare       2017.05.07  Rollback to boot from Slot 1. Small adjustment of NMI on another slot. Just an approximation. Lack M1 signal and speed.
* Dandare       2017.05.08  Using XC8 1.42 compiler
* Dandare       2017.05.14  Fixed Autodetection for divide update. Bug was there since the beginning but only showed with the new timings. -> RELEASE 6.5
* Dandare       2017.09.22  Small change in boot code to allow for locked commands on internal boot rom instead of disabled commands -> Sword of Ianna Cartridge production. 
*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*
* ADDITIONAL NOTES: Public Domain
* Signaling banks is done with 1-32 pulses for banks 0-31
*   
**********************************************************************/

#include "config_bits.h"                                                      
#include <xc.h>
#include "interrupts.h"
#include "common.h"
#include "commands.h"

//** <ENTER --codeoffset=0x0200 in XC8 Global Options> */
    
#define Spectrum_Type_Sense PORTAbits.RA5                                       // Sense "extra" reset signal available in +2 / +3
#define Z80_Reset_Pin_RD PORTBbits.RB6                                          // Pin to read reset status
#define ROMCSReq PORTAbits.RA6                                                  // ROMCS Request from another peripheral connected AFTER this one

#define Eeprom_Disable_nCE() LATAbits.LATA7=1; TRISAbits.TRISA7=0
#define Eeprom_Enable_nCE() LATAbits.LATA7=0; TRISAbits.TRISA7=0
#define Disable_Spectrum48k_nROMCS() LATBbits.LATB3=1; TRISBbits.TRISB3=0
#define Enable_Spectrum48k_nROMCS() TRISBbits.TRISB3=1
#define Disable_SpectrumPlus2y3_nROMCS() LATB|=0x38; TRISB&=0xC7                // Force also ROMCS (A25 - latb.b3)
#define Enable_SpectrumPlus2y3_nROMCS() TRISB|=0x38

bool Spectrum48kFlag;                                                           // Software Flag - Detected Spectrum48k in bootloader (False is for +2 / +3 detection)
bool CheckUpStream=true;                                                        // Software Flag - Check RomCS upstream
eeprom unsigned char EEData[256] =                                              // PIC EEprom Memory
{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,32,33};             // Initialize EEprom on programming (SAME VALUES AS BOOTLOADER)


/*************************************************************************
* Initialize PIC HW
*************************************************************************/
void InitHW(void) 
{
  OSCCON  =0b11110000;                                                          // 32Mhz Internal OSC - Datasheet Pag 65
  APFCON0 =0b00000000;                                                          // Allocate Pins - Datasheet Pag 119
  APFCON1 =0;
  WDTCON  = 0;                                                                  // Disable WatchDog Timer and set interval to 1ms - Datasheet Pag 99
  ANSELA  = 0;                                                                  // Disable analog function on PORTA
  ANSELB  = 0;                                                                  // Disable analog function on PORTB
 
  OPTION_REG = 0b00000111;                                                      // Enable WPUx, INT on Falling edge, PreScaler 1:256 Timer 0 - Datasheet Pag 176
  WPUA  = 0b00100000;                                                           // Enable pullups on A for RA5 - Top 28
  WPUB  = 0b00000111;                                                           // Enable pullups on B for RB0,RB1,RB2 - INT, Switch1, Switch2

  LATAbits.LATA7 = 1;                                                           // Ensure Eeprom is disabled when TRIS to output
  TRISA = 0b01100000;                                                           // #CE, ROMCSReq, Top28, Eeprom A18 - Eeprom A14
  TRISB = 0b11111111;                                                           // NMI, Reset, OE2, OE1, RomCS, SW2, SW1, INT
  
  INTE  = 0;     /*POTENTIAL CLEANUP*/                                          // Disable INT for Spectrum-driven Eeprom Bank Change
  IOCBP = 0b00000010;                                                           // Enable Positive Edge for RB1 -> Switch1 (Positive edge is release button)
  IOCBN = 0b00000110;                                                           // Enable Negative Edge for RB1,RB2 -> Switch1, Switch2 (negative edge is push button)
  IOCIE = 1;                                                                    // Enable IOC for Push Buttons
  INTF = 0;     /*POTENTIAL CLEANUP*/                                           // Clear INT Flag
  IOCBF = 0;    /*POTENTIAL CLEANUP*/                                           // Clear Individual IOCIF registers (IOCIF is automatically cleared if IOCBF=0)
  /* CLEANUP 
  T1CONbits.T1CKPS1 = 1;                                                        // bits 5-4  Prescaler Rate Select bits
  T1CONbits.T1CKPS0 = 1;                                                        // bit 4
  T1CONbits.TMR1CS = 0;                                                         // bit 2 Timer1 Clock Source Select bit...0 = Internal clock (FOSC/4)
  T1CONbits.TMR1ON = 1;                                                         // bit 0 enables timer
  T2CON |= 120;                                                                 // bits 6-3 Post scaler 1:1 thru 1:16
  T2CONbits.TMR2ON = 1;                                                         // bit 2 turn timer2 on;
  T2CONbits.T2CKPS1 = 1;                                                        // bits 1-0  Prescaler Rate Select bits
  T2CONbits.T2CKPS0 = 0;  
  */
  T1CON=0b110001;
  T2CON=0b1111110;
  PEIE = 1;                                                                     // Enable Peripheral Interrupt
  GIE = 0;    /*POTENTIAL CLEANUP*/                                             // Disable Interrupts - Already disabled on boot                                                                              
}



/*************************************************************************
* Enable Dandanator Commands
*************************************************************************/
void Enable_Dandanator_Commands()  
  {
    SpectrumCommandStatus=Wait_Normal;
    ModifyEepromFlag=false;
    ReadEepromFlag=false;
    INTF=0;
    INTE=1;
  }



/*************************************************************************
* Restore Internal Spectrum ROM
*************************************************************************/
void Restore_Internal_ROM(bool reset)
{
   if (reset) 
   {
     Z80_nReset_Hold();                                                         // Put Z80 in reset  
     ChangeSlotOnReset=false;                                                   // Clear Change Slot on Reset
   }
   Eeprom_Disable_nCE();                                                        // Disable External ROM
   if (Spectrum48kFlag)                                                         // Enable Internal ROM depending on Spectrum type
   {
     Enable_Spectrum48k_nROMCS(); 
   }
   else
   {
    Enable_SpectrumPlus2y3_nROMCS();
   }
   if (reset) 
   {
     Z80_nReset_Release();                                                      // Release Z80 from reset      
   }
}



/*************************************************************************
* Select Eeprom Bank
*************************************************************************/
void Select_Eeprom_Bank(unsigned char BankN031)
{
unsigned char ShadowLATA;                                                       // PIC16 infamous RMW problem workaround (Don't know for sure if this part has it)

  ShadowLATA = LATA;                                                            // Get LATA Status
  ShadowLATA&=0xE0;                                                             // Clear lower 5 bits
  ShadowLATA|=(BankN031 & 0x1F);                                                // Add BankN 0-31
  LATA=ShadowLATA;                                                              // Set LATA
}


/*************************************************************************
* Switch to Eeprom Bank
*************************************************************************/
void Switch_To_Eeprom_Bank(unsigned char BankN031,  bool reset)
{
   if (BankN031<(InternalROMBank-1))
   {
     if (reset) 
     {
       Z80_nReset_Hold();                                                       // Put Z80 in reset  
       ChangeSlotOnReset=false;                                                 // Clear Change Slot on Reset
     }
     if (Spectrum48kFlag)                                                       // Disable Internal ROM depending on Spectrum type
     {
       Disable_Spectrum48k_nROMCS(); 
     }
     else
     {
      Disable_SpectrumPlus2y3_nROMCS();
     }
     Eeprom_Enable_nCE();                                                       // Enable External ROM
     Select_Eeprom_Bank(BankN031);                                              // Select specific bank
     if (BankN031==(DandanatorMiniFWBank-1)) Enable_Dandanator_Commands();      // Always Force Commands ON for BANK 0 (Dandanator Mini FW)
     
     if (reset) 
     {
       Z80_nReset_Release();
     }
   }
   else
   {
     Restore_Internal_ROM(reset);
   }
}



/*************************************************************************
* Initialize Switches Behaviour
*************************************************************************/
void Init_Switches()
{
  Switches[SW1][Short].action=RESET;                                            // Default Short Switch1 press is Reset
  Switches[SW1][Short].BankN=0;                                                 // Bank for Reset is unused 
  Switches[SW1][Long].action=Select_RomN_RESET;                                 // Return to Internal ROM
  Switches[SW1][Long].BankN=InternalROMBank;
  Switches[SW1][Double].action=Select_RomN_RESET;                               // Launch Dandanator ROM and reset                             
  Switches[SW1][Double].BankN=DandanatorMiniFWBank;
  
  Switches[SW2][Short].action=Select_RomN_RESET;                                // Launch Dandanator ROM and reset
  Switches[SW2][Short].BankN=DandanatorMiniFWBank;                              // 
//  Switches[SW2][Long].action=Switch_Disabled;                                   // Disabled
//  Switches[SW2][Long].BankN=0;
//  Switches[SW2][Double].action=Switch_Disabled;                                 // Disabled
//  Switches[SW2][Double].BankN=0;
 
  IOCBF=0;                                                                      // Clear Switch Presses Interrupt Flag
}



/*************************************************************************
* Act on Switch
*************************************************************************/
Perform_Switch_Action(unsigned char SWN, unsigned char ACTN)
{
unsigned char tmpSW;
unsigned char tmpShadowLATA;
unsigned char tmpShadowTRISB;
    CheckUpStream=true;                                                         // Re-enable upstream check for ROMCS of any upstream peripheral
    switch (Switches[SWN][ACTN].action)
    {
        case Select_RomN_RESET:
        {
            if (Switches[SWN][ACTN].BankN<InternalROMBank)
            {
                Switch_To_Eeprom_Bank(Switches[SWN][ACTN].BankN-1,true);
                if ((EEData[EE_NormalBoot]!=DandanatorMiniFWBank) && (Switches[SWN][ACTN].BankN==DandanatorMiniFWBank))
                  EEData[EE_NormalBoot]=DandanatorMiniFWBank;                   // Boot will always be slot 1
            }
            else
            {
                if (EEData[EE_NormalBoot]!=InternalROMBank) 
                  EEData[EE_NormalBoot]=InternalROMBank;
                Disable_Dandanator_Commands();
                ChangeSlotOnReset=false;                                        // Clear Change Slot on Spectrum-Driven-Reset
                Restore_Internal_ROM(true);                                     // Switch to internal ROM and reset
                //WDTCONbits.SWDTEN=1;                                          // Enable watch Dog to reset PIC
               // while(1);                                                     // Status.nto not working now???? DEBUG - Apparent bug in Compiler
            }
            break;
        }
        case Select_RomN_NMI_Enable_Commands:                                   // Disabled  check for Internal ROM for the time being. Missing flash space
        {       
//          if (Switches[SWN][ACTN].BankN<InternalROMBank)                      // Disable check for NMI on Internal ROM. WARNING! - Always do NMI/Enable commands on exteral slots
//            {
                if (LATAbits.LATA7==1)                                          // Get current ROM (either internal or external)
                    PreviousSlot = 33;
                else
                    PreviousSlot = (LATA&0x1F)+1;
     
                LATBbits.LATB7=0;                                               // Prepare NMI for fast but not too-fast action (3,5us sweet spot) between NMI and Rom Ready
                tmpSW=Switches[SWN][ACTN].BankN-1;                              // Get Slot Number
                tmpShadowLATA = LATA;                                           // Get LATA Status (eeprom Addresses and CE)
                tmpShadowLATA&=0xE0;                                            // Clear lower 5 bits
                tmpShadowLATA|=(tmpSW & 0x1F);                                  // Add BankN 0-31       
                tmpShadowLATA&=0x7F;                                            // Prepare CE
                tmpShadowTRISB= TRISB;                                          // Get LATB Status (ROMCS OE1 OE2)
                tmpShadowTRISB|=0x38;                                           // Clear Any active ROMCS OE1 or OE2             
                if (Spectrum48kFlag)
                {
                  tmpShadowTRISB&=0xF7;                                         // Set bit 3 to 0 -> Output 
                }
                else
                {
                  tmpShadowTRISB&=0xC7;                                         // Set bits 3,4,5 to 0 -> Output
                }
                TRISBbits.TRISB7=0;                                             // Send NMI
                __delay_us(2);                                                  // Total delay of 3,46us between NMI edge and ROM Enable 
                NOP();
                NOP();
                NOP();
                NOP();
                TRISB=tmpShadowTRISB;                                           // Already releases NMI (this is 4 asm instructions, 500ns) - ALso releases NMI
                LATA=tmpShadowLATA;                                             // Set ROM and Enable CE (this is 4 asm instructions, 500ns))
               // TRISBbits.TRISB7=1;                                             // Release NMI
                Enable_Dandanator_Commands();
 //           }
 //           else
 //           {
 //             ChangeSlotOnReset=false;                                          // Clear Change Slot on Spectrum-Driven-Reset
 //             Restore_Internal_ROM(true);
 //           }
            break;  
        }
        case RESET:
        {
            if (ChangeSlotOnReset)                                              // If spectrum-driven reset is detected
            {
              ChangeSlotOnReset=false;                                          // Disable further changeSlots actions until new command 39 is received
              Switch_To_Eeprom_Bank(ResetSlot,false);                           // ResetSlot will either select an internal bank or external with commands disabled.
            }
            Z80_Send_Reset(); 
            break;
        }
        case NMI:
        {
            Z80_Send_NMI();
            break;
        }
        /* DISABLED TO SAVE SPACE - NOT USED FUNCTIONS
        case Lock_Commands:
        {
            SpectrumCommandStatus=Commands_Locked;
            break; 
        }
        case Disable_Commands:
        {
            Disable_Dandanator_Commands();
            break;
        }
         */
        case Select_RomN_RESET_Enable_Unlock_Commands:
        {
            ChangeSlotOnReset=false;                                            // Clear Change Slot on Spectrum-Driven-Reset
            Switch_To_Eeprom_Bank(Switches[SWN][ACTN].BankN-1,true);
            Enable_Dandanator_Commands();
            break;
        }
        case Next_Rom_RESET:
        {
           if (Switches[SW1][Double].BankN > 1) Switches[SW1][Double].BankN--;  // Move var to previous slot (Roms are stored counting down from slot 32)
            ChangeSlotOnReset=false;                                            // Clear Change Slot on Spectrum-Driven-Reset
            Switch_To_Eeprom_Bank(Switches[SWN][ACTN].BankN-1,true);            // Launch updated Eeprom Bank
            Enable_Dandanator_Commands();                                       // Enable commands
            SpectrumCommandStatus=Commands_Locked;                              // Lock Commands
        }
        default: break;
    }
}



/*************************************************************************
* Boot Spectrum Based on stored info and switch presses
*************************************************************************/
void Boot_Spectrum()
{
unsigned char SelectedBootRom;
 if (ROMCSReq == 0)                                                             // Only boot if no Upstream ROMCSReq is active
 {    
  if ((Switch1==0) && (STATUSbits.nTO==1))                                      // Check for Forced Boot on switch 1 (Only on Cold Boot, not on WDT reset)
  {                                                                             // Status.nto not working now??? DEBUG
    SelectedBootRom=EEData[EE_ForcedROM_Switch1];  
    Switches[SW1][Double].action=Next_Rom_RESET;                                // Assign double click on left button to Change rom and reset                             
    Switches[SW1][Double].BankN=SelectedBootRom;
  }
  else
  {
    if ((Switch2==0) && (STATUSbits.nTO==1))                                    // Check for Forced Boot on switch 2 (Only on Cold Boot, not on WDT reset)
    {                                                                           // Status.nto not working now??? DEBUG
      SelectedBootRom=EEData[EE_ForcedROM_Switch2];  
      if (SelectedBootRom==InternalROMBank)                                     // Enable commands but lock them if booting with button to internal ROM
      {                                                                         // This will allow Command 48 from tape loaded program
        Enable_Dandanator_Commands();
        SpectrumCommandStatus=Commands_Locked;  
        CheckUpStream=false;                                                    // Do not check Upstream RomCS
      }
    }   
    else
    {
      if (EEData[EE_NormalBoot] < InternalROMBank)                              // Check normal boot
      {
        SelectedBootRom=EEData[EE_NormalBoot];    
      }  
      else
      {
        SelectedBootRom=0xFF;                                                   // Dummy value
        // WARNING - SWORD OF IANNA CHANGE 2017.09.22 // Disable_Dandanator_Commands();                                          // Ensure Dandanator does not respond to commands      
            Enable_Dandanator_Commands();
            SpectrumCommandStatus=Commands_Locked;                              // WARNING - SWORD OF IANNA CHANGE 2017.09.22
      }
    }
  }
  if (SelectedBootRom==ExtraROMBank)                                            // Special Case for ExtraROM slot. Commands Enabled&Locked by Default
  {
    Enable_Dandanator_Commands();
    SpectrumCommandStatus=Commands_Locked;
  }
  if ((SelectedBootRom==EEData[EE_ExtraRom_CMDAllowed1]) || (SelectedBootRom==EEData[EE_ExtraRom_CMDAllowed2]))
    Enable_Dandanator_Commands();                                               // Allow Commands on selected Boot Rom if indicated. ROM1, Slot 0 Always enables commands
  if (SelectedBootRom<InternalROMBank)
  {
    Z80_nReset_Hold();   
    Switch_To_Eeprom_Bank(SelectedBootRom-1,false);                             // Switch to that ROM while holding reset
     /* CLEANUP
     INTF=0;                                                                     // Enable Interrupts (already done a bit after)
    IOCBF=0;
    IOCIF=0;
    TMR0IF=0;
    TMR1IF=0;
    TMR2IF=0;
    GIE=1; */
    __delay_ms(Reset_Hold_Time_ms);
    Z80_nReset_Release();                                                       // Release Reset (DEBUG - Long reset hold means flashing squares on screen)
  }
 } 
 else
 {
   Enable_Dandanator_Commands();                                                // If RomCS or Reset are driven, enable commands and lock them (easier divide upgrade and no conflict dandanator stacking)
   SpectrumCommandStatus=Commands_Locked;    
   CheckUpStream=false;
 }
  INTF=0;                                                                       // Enable Interrupts
  IOCBF=0;
  IOCIF=0;
  TMR0IF=0;
  TMR1IF=0;
  TMR2IF=0;
  GIE=1;
}



/*************************************************************************
**************************************************************************
* Main Program
**************************************************************************
*************************************************************************/
void main(void)
{
unsigned char NMICount;                                                         // NMI Pulses responding to Eeprom Read Command
unsigned char NMITop;                                                           
unsigned char ResetTCount;                                                      // Count for Reset-Top28 line stabilization
  InitHW();                                                                     // Init PIC Hardware
  Eeprom_Disable_nCE();                                                         // Disable external Eeprom/Flash
  if (Z80_Reset_Pin_RD==0)                                                      // Wait here until reset line is "1" so Z80 is not in reset and TOP28 is ready for Spectrum Detection
  {
    for(ResetTCount=0;ResetTCount!=150;ResetTCount++)                           // Allow Reset level to set. Top28 has a ~90ms delay from Reset on a +2A
    {
      __delay_ms(1);                                                            // Wait 1ms up to 150 ms to ensure correct Spectrum type detection  
      if(Z80_Reset_Pin_RD==0) ResetTCount=0;                                    // initialize wait count if reset button is still depressed 
    }
  }
  Spectrum48kFlag = (Spectrum_Type_Sense == 1);                                 // Check RESET signal (TOP 28 edge should be 0 for +2/+3 if not in reset- Should be NC for Spectrum 48k->1 (WPUA))
  Init_Switches();                                                              // Initialize default switches behavior 
  Boot_Spectrum();                                                              // Boot based on Stored options
  if (false) Interrupt_Handler();                                               // DEBUG - Force XC8 Optimizer to include Handler - Condition will never be true 
  while (true)                                                                  // Main Loop
  {
    if ((ROMCSReq == 1) && (CheckUpStream==true))                               // Check for another Upstream peripheral trying to get the ROM Address space if no right button boot
    {
      Restore_Internal_ROM(false);                                              // Lock commands, buttons and disable external ROM if upstream peripheral sets ROMCS to 1
      SpectrumCommandStatus=Commands_Locked;                                    // Lock Commands                                                  
      ShortSwitchPressedFlag=false;                                             // Discard all button presses
      DoubleSwitchPressedFlag=false;
      LongSwitchPressedFlag=false;
      Switch2PressedFlag=false;
      ChangeSlotOnReset=false;                                                  // Disable further changeSlots actions until new command is received
    }
    else                                                                        // If Dandanator Mini is allowed to function (no upstream conflict)
    {      
      if (ShortSwitchPressedFlag)                                               // SHORT SWITCH 1 PRESS
      {
        ShortSwitchPressedFlag=false;  
        Perform_Switch_Action(SW1,Short);
      }     
      if (LongSwitchPressedFlag)                                                // LONG SWITCH 1 PRESS
      {  
        LongSwitchPressedFlag=false;
        Perform_Switch_Action(SW1,Long);                                            
      }
      if (DoubleSwitchPressedFlag)                                              // DOUBLE SWITCH 1 PRESS
      {
        DoubleSwitchPressedFlag=false;
        Perform_Switch_Action(SW1,Double);
      }
      if (Switch2PressedFlag)                                                   // SHORT SWITCH 2 PRESS 
      {
        Switch2PressedFlag=false;
        Perform_Switch_Action(SW2,Short);
      }
      if (SpectrumCommandStatus==Pending_Processing)                            // NORMAL COMMAND PENDING PROCESSING
      {
          switch (SpectrumCommandCounter)
          {
             case CMD_RESET:
              {
                  Z80_Send_Reset();
                  Enable_Dandanator_Commands();                                 // Re-enable command reception     
                  break;
              }
              case CMD_NMI:
              {
                  Z80_Send_NMI();
                  Enable_Dandanator_Commands();                                 // Re-enable command reception     
                  break;
              }
              case InternalROMBank:
              {
                  Restore_Internal_ROM(false);                                  // Warning!! Internal ROM with Commands enabled
                  Enable_Dandanator_Commands();                                 // Re-enable command reception     
                  break;
              }
              case CMD_INTERNALROM_DISABLECMD:
              {
                  Restore_Internal_ROM(false);                                  // Internal ROM with Commands disabled
                  Disable_Dandanator_Commands();
                  break;
              }     
              
              case CMD_SET_RESETSLOT:
              {
                  if (LATAbits.LATA7==0)                                        // If external ROM is enabled
                    ResetSlot=LATA&31;                                          // Get current Slot as reset Slot (0-31)
                  else
                    ResetSlot=33;                                               // Reset Slot is 33  (internal commands disabled in 0-33 range)
                  ChangeSlotOnReset=true;                                       // Enable ChangeSlotOnReset
                  Enable_Dandanator_Commands();                                 // Re-enable command reception  
                  break;
              }
              default:                                                          // Change external Rom Slot
              {
                if ((SpectrumCommandCounter>0)&&(SpectrumCommandCounter<InternalROMBank))
                {
                   if  (LATAbits.LATA7==0)                                      // if External Eeprom is already connected
                     Select_Eeprom_Bank(SpectrumCommandCounter-1);
                   else
                     Switch_To_Eeprom_Bank(SpectrumCommandCounter-1,false);     // Enable external rom and switch to it
                }
                Enable_Dandanator_Commands();                                   // Re-enable command reception     
                break;
              };
          }                                              
      }
      if (ModifyEepromFlag)
      {
          ModifyEepromFlag=false;
          if (EEData[EepromAddress]!=ModifyEepromValue)
            EEData[EepromAddress]=ModifyEepromValue;
          Enable_Dandanator_Commands();                                         // Re-enable command reception        
      }
      if (ReadEepromFlag)
      {
          ReadEepromFlag=false;
          NMITop=EEData[EepromAddress]+1;
          for (NMICount=0;NMICount!=NMITop;NMICount++)                          // Send 1-256 pulses for values 0-255
          {
              Z80_Send_NMI();
              __delay_us(20);                                                   // Separate Pulses 20 us
          }
          Enable_Dandanator_Commands();                                         // Re-enable command reception 
      }
      if ((nResetPort==0) & ChangeSlotOnReset)                                  // If spectrum-driven reset is detected
      {
          ChangeSlotOnReset=false;                                              // Disable further changeSlots actions until new command is received
          Switch_To_Eeprom_Bank(ResetSlot,false);                               // ResetSlot will either select an internal bank or external with commands disabled.
      }
    } // End ROMCSReq == 0 branch
  } // End Main Loop
}
 
