/**********************************************************************
* 2016 Dandare: Bootloader of Controller for ZX Dandanator! Mini
* Version:        	2016.11.04
* uController:     	PIC16F1826
* REVISION HISTORY:
*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
* Author        Date        Comments on this revision
*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
* Dandare       2016.07.09  Working Version
* Dandare       2016.07.10  Include Main Program as Loadable - Reset Eeprom on call
* Dandare       2016.08.11  Remove Z80 reset after PIC FW Upgrade
* Dandare       2016.11.04  Compiler aid eeprom initialization with the correct values. 
*                           Does not affect compiled code so same version of bootloader.
*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*
* ADDITIONAL NOTES: Public Domain 
*                   Tested on:  48k, 128k, +2, +2A, +3 and Harlequin Rev.D,F&G 
 *                              Inves, Timex, TK90, TK95, JS128
*                               running 6.0 and higher Z80ASM Dandanator Menu
*   
**********************************************************************/


#include "config_bits.h"
#include "flash.h"
#include <stdbool.h>

#define _XTAL_FREQ 32000000

#define Switch1 PORTBbits.RB1                                                   // Switch 1
#define Switch2 PORTBbits.RB2                                                   // Switch 2

#define Reset_Hold_Time_ms 500                                                  // Time to hold Spectrum in Reset (safe time, longer than usual)
#define Z80_nNMI_Enable() LATBbits.LATB7=0;TRISBbits.TRISB7=0
#define Z80_nNMI_Disable() TRISBbits.TRISB7=1
#define Z80_Send_NMI() Z80_nNMI_Enable();NOP();NOP();Z80_nNMI_Disable()
#define Z80_nReset_Hold() LATBbits.LATB6=0;TRISBbits.TRISB6=0;__delay_ms(Reset_Hold_Time_ms) // ensure minimum time for reset pulse
#define Z80_nReset_Release() TRISBbits.TRISB6=1
#define Z80_Send_Reset() Z80_nReset_Hold();Z80_nReset_Release()

#define Spectrum_Type_Sense PORTAbits.RA5                                       // Sense "extra" reset signal available in +2 / +3

#define Eeprom_Disable_nCE() LATAbits.LATA7=1; TRISAbits.TRISA7=0
#define Eeprom_Enable_nCE() LATAbits.LATA7=0; TRISAbits.TRISA7=0
#define Disable_Spectrum48k_nROMCS() LATBbits.LATB3=1; TRISBbits.TRISB3=0
#define Enable_Spectrum48k_nROMCS() TRISBbits.TRISB3=1
#define Disable_SpectrumPlus2y3_nROMCS() LATB|=0x30; TRISB&=0xCF
#define Enable_SpectrumPlus2y3_nROMCS() TRISB|=0x30

#define EE_ExtraRom_CMDAllowed1 251                                             // Eeprom Addresses
#define EE_ExtraRom_CMDAllowed2 252
#define EE_NormalBoot 253                                                       
#define EE_ForcedROM_Switch1 254
#define EE_ForcedROM_Switch2 255

#define HE_0 0x44                                                               // Firmware Upgrade 10-byte Header =  "DNTRMFW-Up"
#define HE_1 0x4E
#define HE_2 0x54
#define HE_3 0x52
#define HE_4 0x4D
#define HE_5 0x46
#define HE_6 0x57
#define HE_7 0x2D
#define HE_8 0x55
#define HE_9 0x70

bool firstbyte=true;

eeprom unsigned char EEData[256] =                                              // PIC EEprom Memory
{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,32,33};             // Initialize EEprom on programming (SAME VALUES AS MAIN CODE)

/*************************************************************************
* Initialize Hardware
*************************************************************************/
void Init_Remaining_HW(void) 
{
  APFCON0 =0b00000000;                                                          // Allocate Pins - Datasheet Pag 119
  APFCON1 =0;
  WDTCON  = 0;                                                                  // Disable WatchDog Timer and set interval to 1ms - Datasheet Pag 99
  ANSELA  = 0;                                                                  // Disable analog function on PORTA

  WPUA  = 0b00100000;                                                           // Enable pullups on A for RA5 - Top 28

  LATAbits.LATA7 = 1;                                                           // Ensure Eeprom is disabled when TRIS to output
  TRISA = 0b01100000;                                                           // #CE, ROMCSReq, Top28, Eeprom A18 - Eeprom A14
  TRISB = 0b11111111;                                                           // NMI, Reset, OE2, OE1, RomCS, SW2, SW1, INT
  
  GIE = 0;                                                                      // Disable Interrupts
}



/*************************************************************************
* ISR Redirection
*************************************************************************/
void interrupt ISR(void)
{
  asm("call " ___mkstr(FLASHSTORAGEPOSITION + ISR_OFFSET));                     // Jump to main program ISR Routine
}



/*************************************************************************
* Boot ZX Dandanator Mini
*************************************************************************/
void Boot_ZX_Dandanator_Mini(void)
{
  asm("goto " ___mkstr(FLASHSTORAGEPOSITION));                                  // Jump to main program
}



/*************************************************************************
* Receive Byte from Spectrum in a set of MREQ WR to ROM pulses
*************************************************************************/
unsigned char Receive_Byte(void)
{
#define MAXLOOPS 255                                                            // Number of loops with 0 pulses from Spectrum to determine end of byte
#define MAXLOOPSFIRST 65000
unsigned short nloop;
unsigned char cnt;
unsigned short WaitLoops;
  if (firstbyte)                                                                // After First NMI, Spectrum must load new screen and move data to RAM
  { 
    WaitLoops=MAXLOOPSFIRST;                                                    // So we wait longer
    firstbyte=false;
  }
  else
  { 
    WaitLoops=MAXLOOPS;
  }
  nloop=0;
  cnt=0;    
  INTF=0;                                                                       // Clear INT Interrupt Flag. Interrupts are disabled. Will check for active Flag
  Z80_Send_NMI();                                                               // Send NMI to Spectrum (NMI ISR will return next byte)
  
  while (nloop<WaitLoops)
  {
      nloop++;
      if (INTF==1)                                                              // If pulse is received from Spectrum
      {
          nloop=0;                                                              // Clear nloops
          cnt++;                                                                // Increment Counter
          WaitLoops=MAXLOOPS;                                                   // Reset first wait
          INTF=0;                                                               // Clear Int Flag
      }
  }
  return cnt;                                                                   // Return received number of pulses
}




/*************************************************************************
* Main Bootloader
*************************************************************************/
void main(void) 
{
#define HEADERSIZE 10                                                           // Size of flash upgrade header
const unsigned char Header[HEADERSIZE]={HE_0,HE_1,HE_2,HE_3,HE_4,HE_5,HE_6,HE_7,HE_8,HE_9}; // Header of Firmware
unsigned short ChunkCnt;                                                        // Flash chunk counter
unsigned short SegmentCnt;
unsigned char ByteCnt;                                                          // Flash byte counter within chunk
bool HeaderOk;                                                                  // Header validation var   
bool Spectrum48kFlag;                                                           // HW Detection of Spectrum type
  OSCCON  =0b11110000;                                                          // 32Mhz Internal OSC - Datasheet Pag 65
  ANSELB  = 0;                                                                  // Disable analog function on PORTB (TRISB is 0xFF on boot)
  OPTION_REG = 0b00000111;                                                      // Enable WPUx, INT on Falling edge, PreScaler 1:256 Timer 0 - Datasheet Pag 176
  WPUB  = 0b00000111;                                                           // Enable pullups on B for RB0,RB1,RB2 - INT, Switch1, Switch2
  __delay_ms(40);                                                               // Wait Pull-ups to charge external 100nf capacitors.
  if ((Switch1==0) && (Switch2==0))                                             // If both switches -> Enable Firmware Upgrade
  {
    Init_Remaining_HW();                                                        // Initialize uninitialized PIC Hardware
    // Boot to First External Rom Slot
    Eeprom_Disable_nCE();                                                       // Disable external Eeprom/Flash
    __delay_ms(10);                                                             // Allow Reset level to set and Spectrum to boot - Warning don't update pressing reset as Top28 is delayed ~90ms on a +2A
    Spectrum48kFlag = (Spectrum_Type_Sense == 1);                               // Check RESET signal (TOP 28 edge should be 0 for +2/+3 if not in reset- Should be NC for Spectrum 48k->1 (WPUA))
    Z80_nReset_Hold();  
    if (Spectrum48kFlag)                                                        // Disable Internal ROM depending on Spectrum type
    {
       Disable_Spectrum48k_nROMCS(); 
    }
    else
    {
      Disable_SpectrumPlus2y3_nROMCS();
    }
    Eeprom_Enable_nCE();                                                        // Enable External ROM
    LATA=0;                                                                     // Fast move to Slot 0 and enable CE as well
    Z80_nReset_Release();
    EEData[EE_ExtraRom_CMDAllowed1]=1;                                          // Initialize Eeprom Parameters
    EEData[EE_ExtraRom_CMDAllowed2]=1;
    EEData[EE_NormalBoot]=1;                                                    // Force Normal boot into slot 1
    EEData[EE_ForcedROM_Switch1]=32;                                            // Force Switch1 boot as last slot                                        
    EEData[EE_ForcedROM_Switch2]=33;                                            // Force Switch2 boot as internal ROM
    __delay_ms(400);                                                            // Allow Spectrum to boot in Slot 0 (safe time)
    
    // Receive Header
    ByteCnt = 0;                                                                // Init ByteCnt
    HeaderOk = true;                                                            // Init HeaderOk to ok (for the moment)
    while ((ByteCnt<HEADERSIZE) && (HeaderOk))
    {
        if (Receive_Byte()!=Header[ByteCnt]) HeaderOk = false;                  // If header is not the same, abort
        ByteCnt++;
    }
     
    // Receive Firmware Payload     
    if (HeaderOk)                                                               // Received Header is Ok
    {
      for (SegmentCnt=0;SegmentCnt!=NSEGMENTS;SegmentCnt++)                     // Erase All Segments
        Erase_Flash_Segment(SegmentCnt);
      for (ChunkCnt=0;ChunkCnt!=NCHUNKS;ChunkCnt++)                             // All Chunks
      {
          for (ByteCnt=0;ByteCnt!=(FLASHCHUNKSIZE*2);ByteCnt++)                 // All Bytes in a Chunk
          {
              FlashBuffer[ByteCnt]=Receive_Byte();                              // Populate flashBuffer
          }
          Save_Flash_Chunk(ChunkCnt,FlashBuffer);                               // Save to Flash
      }
      SWDTEN=1;                                                                 // Reset PIC
      while(1);
    }
    else                                                                        // Received Header is not Ok
    {
      while ((Switch1==0) || (Switch2==0));                                     // Wait for button release
      LATA=0xFF;                                                                // Disable eeprom
      TRISA=0xFF;
      TRISB=0xFF;                                                               // Re-enable Internal ROM
      SWDTEN=1;                                                                 // Reset PIC
      while(1);
    }
  }
  Boot_ZX_Dandanator_Mini();                                                    // Boot Dandanator in any case
}
